@@ -29,7 +29,6 @@ import tempfile
 import shutil
 
 from firewall.core.logger import log
-from firewall.functions import b2u, u2b, PY2
 
 class ifcfg(object):
     def __init__(self, filename):
@@ -49,8 +48,8 @@ class ifcfg(object):
         return self._config.get(key.strip())
 
     def set(self, key, value):
-        _key = b2u(key.strip())
-        self._config[_key] = b2u(value.strip())
+        _key = key.strip()
+        self._config[_key] = value.strip()
         if _key in self._deleted:
             self._deleted.remove(_key)
 
@@ -60,7 +59,7 @@ class ifcfg(object):
             if s:
                 s += '\n'
             s += '%s=%s' % (key, value)
-        return u2b(s) if PY2 else s
+        return s
 
     # load self.filename
     def read(self):
@@ -80,9 +79,11 @@ class ifcfg(object):
             # get key/value pair
             pair = [ x.strip() for x in line.split("=", 1) ]
             if len(pair) != 2:
-                log.warning("%: Invalid option definition: '%s'", self.filename, line.strip())
                 continue
-            elif pair[1] == '':
+            if len(pair[1]) >= 2 and \
+               pair[1].startswith('"') and pair[1].endswith('"'):
+                pair[1] = pair[1][1:-1]
+            if pair[1] == '':
                 continue
             elif self._config.get(pair[0]) is not None:
                 log.warning("%s: Duplicate option definition: '%s'", self.filename, line.strip())
@@ -98,9 +99,6 @@ class ifcfg(object):
         # handled keys
         done = [ ]
 
-        #if not os.path.exists(ETC_FIREWALLD):
-        #    os.mkdir(ETC_FIREWALLD, 0o750)
-
         try:
             temp_file = tempfile.NamedTemporaryFile(
                 mode='wt',
@@ -143,6 +141,9 @@ class ifcfg(object):
                         continue
                     key = p[0].strip()
                     value = p[1].strip()
+                    if len(value) >= 2 and \
+                       value.startswith('"') and value.endswith('"'):
+                        value = value[1:-1]
                     # check for modified key/value pairs
                     if key not in done:
                         if key in self._config and self._config[key] != value:
