@@ -29,7 +29,8 @@ from firewall.functions import tempFile, readfile, splitArgs, check_mac, portStr
 from firewall import config
 from firewall.errors import FirewallError, INVALID_PASSTHROUGH, INVALID_RULE, UNKNOWN_ERROR, INVALID_ADDR
 from firewall.core.rich import Rich_Accept, Rich_Reject, Rich_Drop, Rich_Mark, \
-                               Rich_Masquerade, Rich_ForwardPort, Rich_IcmpBlock
+                               Rich_Masquerade, Rich_ForwardPort, Rich_IcmpBlock, Rich_Tcp_Mss_Clamp
+from firewall.core.base import DEFAULT_ZONE_TARGET
 import string
 
 POLICY_CHAIN_PREFIX = ""
@@ -295,10 +296,7 @@ class ip4tables(object):
 
                 index = zone_source_index_cache.index(zone_source)
             else:
-                if self._fw._allow_zone_drifting:
-                    index = 0
-                else:
-                    index = len(zone_source_index_cache)
+                index = len(zone_source_index_cache)
 
             rule[0] = "-I"
             rule.insert(2, "%d" % (index + 1))
@@ -534,9 +532,11 @@ class ip4tables(object):
     def _detect_wait_option(self):
         wait_option = ""
         ret = runProg(self._command, ["-w", "-L", "-n"])  # since iptables-1.4.20
+        log.debug3("%s: %s: probe for wait option (%s): ret=%u, output=\"%s\"", self.__class__, self._command, "-w", ret[0], ret[1])
         if ret[0] == 0:
             wait_option = "-w"  # wait for xtables lock
             ret = runProg(self._command, ["-w10", "-L", "-n"])  # since iptables > 1.4.21
+            log.debug3("%s: %s: probe for wait option (%s): ret=%u, output=\"%s\"", self.__class__, self._command, "-w10", ret[0], ret[1])
             if ret[0] == 0:
                 wait_option = "-w10"  # wait max 10 seconds
             log.debug2("%s: %s will be using %s option.", self.__class__, self._command, wait_option)
@@ -551,6 +551,7 @@ class ip4tables(object):
         wait_option = ""
         for test_option in ["-w", "--wait=2"]:
             ret = runProg(self._restore_command, [test_option], stdin=temp_file.name)
+            log.debug3("%s: %s: probe for wait option (%s): ret=%u, output=\"%s\"", self.__class__, self._command, test_option, ret[0], ret[1])
             if ret[0] == 0 and "invalid option" not in ret[1] \
                            and "unrecognized option" not in ret[1]:
                 wait_option = test_option
@@ -646,10 +647,11 @@ class ip4tables(object):
                 self.our_chains["raw"].add("%s_direct" % chain)
 
                 if chain == "PREROUTING":
-                    for dispatch_suffix in ["POLICIES_pre", "ZONES_SOURCE", "ZONES", "POLICIES_post"] if self._fw._allow_zone_drifting else ["POLICIES_pre", "ZONES", "POLICIES_post"]:
+                    for dispatch_suffix in ["POLICIES_pre", "ZONES", "POLICIES_post"]:
                         default_rules["raw"].append("-N %s_%s" % (chain, dispatch_suffix))
-                        default_rules["raw"].append("-A %s -j %s_%s" % (chain, chain, dispatch_suffix))
                         self.our_chains["raw"].update(set(["%s_%s" % (chain, dispatch_suffix)]))
+                    for dispatch_suffix in ["ZONES"]:
+                        default_rules["raw"].append("-A %s -j %s_%s" % (chain, chain, dispatch_suffix))
 
         if self.get_available_tables("mangle"):
             default_rules["mangle"] = [ ]
@@ -660,10 +662,11 @@ class ip4tables(object):
                 self.our_chains["mangle"].add("%s_direct" % chain)
 
                 if chain == "PREROUTING":
-                    for dispatch_suffix in ["POLICIES_pre", "ZONES_SOURCE", "ZONES", "POLICIES_post"] if self._fw._allow_zone_drifting else ["POLICIES_pre", "ZONES", "POLICIES_post"]:
+                    for dispatch_suffix in ["POLICIES_pre", "ZONES", "POLICIES_post"]:
                         default_rules["mangle"].append("-N %s_%s" % (chain, dispatch_suffix))
-                        default_rules["mangle"].append("-A %s -j %s_%s" % (chain, chain, dispatch_suffix))
                         self.our_chains["mangle"].update(set(["%s_%s" % (chain, dispatch_suffix)]))
+                    for dispatch_suffix in ["ZONES"]:
+                        default_rules["mangle"].append("-A %s -j %s_%s" % (chain, chain, dispatch_suffix))
 
         if self.get_available_tables("nat"):
             default_rules["nat"] = [ ]
@@ -674,10 +677,11 @@ class ip4tables(object):
                 self.our_chains["nat"].add("%s_direct" % chain)
 
                 if chain in [ "PREROUTING", "POSTROUTING" ]:
-                    for dispatch_suffix in ["POLICIES_pre", "ZONES_SOURCE", "ZONES", "POLICIES_post"] if self._fw._allow_zone_drifting else ["POLICIES_pre", "ZONES", "POLICIES_post"]:
+                    for dispatch_suffix in ["POLICIES_pre", "ZONES", "POLICIES_post"]:
                         default_rules["nat"].append("-N %s_%s" % (chain, dispatch_suffix))
-                        default_rules["nat"].append("-A %s -j %s_%s" % (chain, chain, dispatch_suffix))
                         self.our_chains["nat"].update(set(["%s_%s" % (chain, dispatch_suffix)]))
+                    for dispatch_suffix in ["ZONES"]:
+                        default_rules["nat"].append("-A %s -j %s_%s" % (chain, chain, dispatch_suffix))
 
         default_rules["filter"] = []
         self.our_chains["filter"] = set()
@@ -686,10 +690,11 @@ class ip4tables(object):
         default_rules["filter"].append("-N INPUT_direct")
         default_rules["filter"].append("-A INPUT -j INPUT_direct")
         self.our_chains["filter"].update(set("INPUT_direct"))
-        for dispatch_suffix in ["POLICIES_pre", "ZONES_SOURCE", "ZONES", "POLICIES_post"] if self._fw._allow_zone_drifting else ["POLICIES_pre", "ZONES", "POLICIES_post"]:
+        for dispatch_suffix in ["POLICIES_pre", "ZONES", "POLICIES_post"]:
             default_rules["filter"].append("-N INPUT_%s" % (dispatch_suffix))
-            default_rules["filter"].append("-A INPUT -j INPUT_%s" % (dispatch_suffix))
             self.our_chains["filter"].update(set("INPUT_%s" % (dispatch_suffix)))
+        for dispatch_suffix in ["ZONES"]:
+            default_rules["filter"].append("-A INPUT -j INPUT_%s" % (dispatch_suffix))
         if log_denied != "off":
             default_rules["filter"].append("-A INPUT -m conntrack --ctstate INVALID %%LOGTYPE%% -j LOG --log-prefix 'STATE_INVALID_DROP: '")
         default_rules["filter"].append("-A INPUT -m conntrack --ctstate INVALID -j DROP")
@@ -704,17 +709,14 @@ class ip4tables(object):
         self.our_chains["filter"].update(set("FORWARD_direct"))
         for dispatch_suffix in ["POLICIES_pre"]:
             default_rules["filter"].append("-N FORWARD_%s" % (dispatch_suffix))
-            default_rules["filter"].append("-A FORWARD -j FORWARD_%s" % (dispatch_suffix))
             self.our_chains["filter"].update(set("FORWARD_%s" % (dispatch_suffix)))
-        for direction in ["IN", "OUT"]:
-            for dispatch_suffix in ["ZONES_SOURCE", "ZONES"] if self._fw._allow_zone_drifting else ["ZONES"]:
-                default_rules["filter"].append("-N FORWARD_%s_%s" % (direction, dispatch_suffix))
-                default_rules["filter"].append("-A FORWARD -j FORWARD_%s_%s" % (direction, dispatch_suffix))
-                self.our_chains["filter"].update(set("FORWARD_%s_%s" % (direction, dispatch_suffix)))
-        for dispatch_suffix in ["POLICIES_post"]:
+        for dispatch_suffix in ["ZONES"]:
             default_rules["filter"].append("-N FORWARD_%s" % (dispatch_suffix))
             default_rules["filter"].append("-A FORWARD -j FORWARD_%s" % (dispatch_suffix))
             self.our_chains["filter"].update(set("FORWARD_%s" % (dispatch_suffix)))
+        for dispatch_suffix in ["POLICIES_post"]:
+            default_rules["filter"].append("-N FORWARD_%s" % (dispatch_suffix))
+            self.our_chains["filter"].update(set("FORWARD_%s" % (dispatch_suffix)))
         if log_denied != "off":
             default_rules["filter"].append("-A FORWARD -m conntrack --ctstate INVALID %%LOGTYPE%% -j LOG --log-prefix 'STATE_INVALID_DROP: '")
         default_rules["filter"].append("-A FORWARD -m conntrack --ctstate INVALID -j DROP")
@@ -725,6 +727,7 @@ class ip4tables(object):
         default_rules["filter"] += [
             "-N OUTPUT_direct",
 
+            "-A OUTPUT -m conntrack --ctstate RELATED,ESTABLISHED -j ACCEPT",
             "-A OUTPUT -o lo -j ACCEPT",
             "-A OUTPUT -j OUTPUT_direct",
         ]
@@ -749,7 +752,7 @@ class ip4tables(object):
 
     def get_zone_table_chains(self, table):
         if table == "filter":
-            return { "INPUT", "FORWARD_IN", "FORWARD_OUT" }
+            return { "INPUT", "FORWARD" }
         if table == "mangle":
             if "mangle" in self.get_available_tables():
                 return { "PREROUTING" }
@@ -786,7 +789,7 @@ class ip4tables(object):
             if ipv in ["ipv4", "ipv6"] and not self.is_ipv_supported(ipv):
                 continue
             # iptables can not match destination MAC
-            if check_mac(addr) and chain in ["POSTROUTING", "FORWARD_OUT", "OUTPUT"]:
+            if check_mac(addr) and chain in ["POSTROUTING", "FORWARD", "OUTPUT"]:
                 continue
 
             egress_fragments.append(self._rule_addr_fragment("-d", addr))
@@ -847,8 +850,7 @@ class ip4tables(object):
             "PREROUTING": "-i",
             "POSTROUTING": "-o",
             "INPUT": "-i",
-            "FORWARD_IN": "-i",
-            "FORWARD_OUT": "-o",
+            "FORWARD": "-i",
             "OUTPUT": "-o",
         }[chain]
 
@@ -897,21 +899,15 @@ class ip4tables(object):
             "PREROUTING": "-s",
             "POSTROUTING": "-d",
             "INPUT": "-s",
-            "FORWARD_IN": "-s",
-            "FORWARD_OUT": "-d",
+            "FORWARD": "-s",
             "OUTPUT": "-d",
         }[chain]
 
-        if self._fw._allow_zone_drifting:
-            zone_dispatch_chain = "%s_ZONES_SOURCE" % (chain)
-        else:
-            zone_dispatch_chain = "%s_ZONES" % (chain)
-
         # iptables can not match destination MAC
-        if check_mac(address) and chain in ["POSTROUTING", "FORWARD_OUT", "OUTPUT"]:
+        if check_mac(address) and chain in ["POSTROUTING", "FORWARD", "OUTPUT"]:
             return []
 
-        rule = [add_del, zone_dispatch_chain, "%%ZONE_SOURCE%%", zone, "-t", table]
+        rule = [add_del, "%s_ZONES" % (chain), "%%ZONE_SOURCE%%", zone, "-t", table]
         rule.extend(self._rule_addr_fragment(opt, address))
         rule.extend(["-g", _policy])
 
@@ -922,6 +918,7 @@ class ip4tables(object):
         add_del_rule = { True: "-A", False: "-D" }[enable]
         isSNAT = True if (table == "nat" and chain == "POSTROUTING") else False
         _policy = self._fw.policy.policy_base_chain_name(policy, table, POLICY_CHAIN_PREFIX, isSNAT=isSNAT)
+        p_obj = self._fw.policy.get_policy(policy)
 
         self.our_chains[table].update(set([_policy,
                                       "%s_log" % _policy,
@@ -937,17 +934,21 @@ class ip4tables(object):
         rules.append([ add_del_chain, "%s_deny" % _policy, "-t", table ])
         rules.append([ add_del_chain, "%s_allow" % _policy, "-t", table ])
         rules.append([ add_del_chain, "%s_post" % _policy, "-t", table ])
+        if p_obj.derived_from_zone:
+            rules.append([ add_del_rule, _policy, "-t", table, "-j", "%s_%s" % (chain, "POLICIES_pre") ])
         rules.append([ add_del_rule, _policy, "-t", table, "-j", "%s_pre" % _policy ])
         rules.append([ add_del_rule, _policy, "-t", table, "-j", "%s_log" % _policy ])
         rules.append([ add_del_rule, _policy, "-t", table, "-j", "%s_deny" % _policy ])
         rules.append([ add_del_rule, _policy, "-t", table, "-j", "%s_allow" % _policy ])
         rules.append([ add_del_rule, _policy, "-t", table, "-j", "%s_post" % _policy ])
+        if p_obj.derived_from_zone:
+            rules.append([ add_del_rule, _policy, "-t", table, "-j", "%s_%s" % (chain, "POLICIES_post") ])
 
         target = self._fw.policy._policies[policy].target
 
         if self._fw.get_log_denied() != "off":
             if table == "filter":
-                if target in [ "REJECT", "%%REJECT%%" ]:
+                if target in [DEFAULT_ZONE_TARGET, "REJECT", "%%REJECT%%" ]:
                     rules.append([ add_del_rule, _policy, "-t", table, "%%LOGTYPE%%",
                                    "-j", "LOG", "--log-prefix",
                                    "\"%s_REJECT: \"" % _policy ])
@@ -957,8 +958,12 @@ class ip4tables(object):
                                    "\"%s_DROP: \"" % _policy ])
 
         if table == "filter" and \
-           target in [ "ACCEPT", "REJECT", "%%REJECT%%", "DROP" ]:
-            rules.append([ add_del_rule, _policy, "-t", table, "-j", target ])
+           target in [DEFAULT_ZONE_TARGET, "ACCEPT", "REJECT", "%%REJECT%%", "DROP" ]:
+            if target in [DEFAULT_ZONE_TARGET]:
+                _target = "REJECT"
+            else:
+                _target = target
+            rules.append([ add_del_rule, _policy, "-t", table, "-j", _target ])
 
         if not enable:
             rules.reverse()
@@ -971,7 +976,7 @@ class ip4tables(object):
         return []
 
     def _rich_rule_chain_suffix(self, rich_rule):
-        if type(rich_rule.element) in [Rich_Masquerade, Rich_ForwardPort, Rich_IcmpBlock]:
+        if type(rich_rule.element) in [Rich_Masquerade, Rich_ForwardPort, Rich_IcmpBlock, Rich_Tcp_Mss_Clamp]:
             # These are special and don't have an explicit action
             pass
         elif rich_rule.action:
@@ -981,7 +986,7 @@ class ip4tables(object):
             raise FirewallError(INVALID_RULE, "No rule action specified.")
 
         if rich_rule.priority == 0:
-            if type(rich_rule.element) in [Rich_Masquerade, Rich_ForwardPort] or \
+            if type(rich_rule.element) in [Rich_Masquerade, Rich_ForwardPort, Rich_Tcp_Mss_Clamp] or \
                type(rich_rule.action) in [Rich_Accept, Rich_Mark]:
                 return "allow"
             elif type(rich_rule.element) in [Rich_IcmpBlock] or \
@@ -1092,15 +1097,22 @@ class ip4tables(object):
             return []
 
         rule_fragment = []
-        if rich_dest.invert:
-            rule_fragment.append("!")
-        if check_single_address("ipv6", rich_dest.addr):
-            rule_fragment += [ "-d", normalizeIP6(rich_dest.addr) ]
-        elif check_address("ipv6", rich_dest.addr):
-            addr_split = rich_dest.addr.split("/")
-            rule_fragment += [ "-d", normalizeIP6(addr_split[0]) + "/" + addr_split[1] ]
-        else:
-            rule_fragment += [ "-d", rich_dest.addr ]
+        if rich_dest.addr:
+            if rich_dest.invert:
+                rule_fragment.append("!")
+            if check_single_address("ipv6", rich_dest.addr):
+                rule_fragment += [ "-d", normalizeIP6(rich_dest.addr) ]
+            elif check_address("ipv6", rich_dest.addr):
+                addr_split = rich_dest.addr.split("/")
+                rule_fragment += [ "-d", normalizeIP6(addr_split[0]) + "/" + addr_split[1] ]
+            else:
+                rule_fragment += [ "-d", rich_dest.addr ]
+        elif rich_dest.ipset:
+            rule_fragment += [ "-m", "set" ]
+            if rich_dest.invert:
+                rule_fragment.append("!")
+            flags = self._fw.zone._ipset_match_flags(rich_dest.ipset, "dst")
+            rule_fragment += [ "--match-set", rich_dest.ipset, flags ]
 
         return rule_fragment
 
@@ -1185,6 +1197,34 @@ class ip4tables(object):
 
         return rules
 
+    def build_policy_tcp_mss_clamp_rules(self, enable, policy, tcp_mss_clamp_value, destination=None, rich_rule=None):
+        table = "filter"
+        _policy = self._fw.policy.policy_base_chain_name(policy, table, POLICY_CHAIN_PREFIX)
+        add_del = { True: "-A", False: "-D" }[enable]
+
+        rule_fragment = []
+        if rich_rule:
+            chain_suffix = self._rich_rule_chain_suffix(rich_rule)
+            rule_fragment += self._rich_rule_priority_fragment(rich_rule)
+            rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
+            rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
+
+        rules = []
+        rule_fragment = ["-p", "tcp"]
+        if tcp_mss_clamp_value == "pmtu" or tcp_mss_clamp_value is None:
+            rule_fragment += ["--tcp-flags", "SYN,RST", "SYN","-j", "TCPMSS", "--clamp-mss-to-pmtu"]
+        else:
+            rule_fragment += ["--tcp-flags", "SYN,RST", "SYN", "-j", "TCPMSS", "--set-mss", tcp_mss_clamp_value]
+
+        if rich_rule:
+            chain_suffix = self._rich_rule_chain_suffix(rich_rule)
+            rule_fragment += self._rich_rule_priority_fragment(rich_rule)
+            rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
+            rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
+        rules.append(["-t", "filter", add_del, "%s_%s" % (_policy, chain_suffix)]
+                     + rule_fragment)
+        return rules
+
     def build_policy_source_ports_rules(self, enable, policy, proto, port,
                                      destination=None, rich_rule=None):
         add_del = { True: "-A", False: "-D" }[enable]
@@ -1266,21 +1306,6 @@ class ip4tables(object):
                      + rule_fragment +
                      [ "!", "-o", "lo", "-j", "MASQUERADE" ])
 
-        rule_fragment = []
-        if rich_rule:
-            chain_suffix = self._rich_rule_chain_suffix(rich_rule)
-            rule_fragment += self._rich_rule_priority_fragment(rich_rule)
-            rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
-            rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
-        else:
-            chain_suffix = "allow"
-
-        table = "filter"
-        _policy = self._fw.policy.policy_base_chain_name(policy, table, POLICY_CHAIN_PREFIX)
-        rules.append(["-t", "filter", add_del, "%s_%s" % (_policy, chain_suffix)]
-                     + rule_fragment +
-                     ["-m", "conntrack", "--ctstate", "NEW,UNTRACKED", "-j", "ACCEPT" ])
-
         return rules
 
     def build_policy_forward_port_rules(self, enable, policy, port,
@@ -1371,7 +1396,7 @@ class ip4tables(object):
         _policy = self._fw.policy.policy_base_chain_name(policy, table, POLICY_CHAIN_PREFIX)
 
         rules = []
-        rule_idx = 6
+        rule_idx = 8
 
         if self._fw.policy.query_icmp_block_inversion(policy):
             ibi_target = "%%REJECT%%"
@@ -1469,7 +1494,7 @@ class ip6tables(ip4tables):
                               "--log-prefix", "\"RFC3964_IPv4_REJECT: \""])
 
         # Inject into FORWARD and OUTPUT chains
-        rules.append(["-t", "filter", "-I", "OUTPUT", "3",
+        rules.append(["-t", "filter", "-I", "OUTPUT", "4",
                       "-j", chain_name])
         rules.append(["-t", "filter", "-I", "FORWARD", "4",
                       "-j", chain_name])
